@@ -725,3 +725,13 @@ values = {
 }
 
 assert sorted(values.values())[1:] == list(range(2, 34))
+
+def assert_equal(x, y):
+    assert x == y
+
+# Check that we can distinguish names from strings in literal collections, e.g. lists.
+# The list has to have more than 4 items to get accumulated in a collection
+a = ["y", 'Exception', "x", Exception, "z"]
+
+assert_equal(a[1], "Exception")
+assert_equal(a[3], Exception)
