@@ -1,4 +1,4 @@
-#  Copyright (c) 2016, 2018-2020 by Rocky Bernstein
+#  Copyright (c) 2016, 2018-2025 by Rocky Bernstein
 #  Copyright (c) 2005 by Dan Pascu <dan@windowmaker.org>
 #  Copyright (c) 2000-2002 by hartmut Goebel <h.goebel@crazy-compilers.com>
 #  Copyright (c) 1999 John Aycock
@@ -21,48 +21,59 @@ scanner/ingestion module. From here we call various version-specific
 scanners, e.g. for Python 2.7 or 3.4.
 """
 
-from __future__ import print_function
-
+import importlib
+from abc import ABC
 from array import array
 from collections import namedtuple
-import sys
+from types import ModuleType
+from typing import Optional, Union
 
-from uncompyle6 import PYTHON3, IS_PYPY
-from uncompyle6.scanners.tok import Token
 import xdis
-from xdis import Bytecode, canonic_python_version, code2num, instruction_size, extended_arg_val, next_offset
+from xdis import (
+    Bytecode,
+    canonic_python_version,
+    code2num,
+    extended_arg_val,
+    instruction_size,
+    next_offset,
+)
+from xdis.version_info import IS_PYPY, version_tuple_to_str
+
+from uncompyle6.scanners.tok import Token
 
 # The byte code versions we support.
-# Note: these all have to be floats
+# Note: these all have to be tuples of 2 ints
 PYTHON_VERSIONS = frozenset(
     (
-        1.0,
-        1.1,
-        1.3,
-        1.4,
-        1.5,
-        1.6,
-        2.1,
-        2.2,
-        2.3,
-        2.4,
-        2.5,
-        2.6,
-        2.7,
-        3.0,
-        3.1,
-        3.2,
-        3.3,
-        3.4,
-        3.5,
-        3.6,
-        3.7,
-        3.8,
-        3.9,
+        (1, 0),
+        (1, 1),
+        (1, 3),
+        (1, 4),
+        (1, 5),
+        (1, 6),
+        (2, 1),
+        (2, 2),
+        (2, 3),
+        (2, 4),
+        (2, 5),
+        (2, 6),
+        (2, 7),
+        (3, 0),
+        (3, 1),
+        (3, 2),
+        (3, 3),
+        (3, 4),
+        (3, 5),
+        (3, 6),
+        (3, 7),
+        (3, 8),
     )
 )
 
-CANONIC2VERSION = dict((canonic_python_version[str(v)], v) for v in PYTHON_VERSIONS)
+CANONIC2VERSION = dict(
+    (canonic_python_version[version_tuple_to_str(python_version)], python_version)
+    for python_version in PYTHON_VERSIONS
+)
 
 # Magic changed mid version for Python 3.5.2. Compatibility was added for
 # the older 3.5 interpreter magic.
@@ -70,19 +81,17 @@ CANONIC2VERSION["3.5.2"] = 3.5
 
 
 # FIXME: DRY
-if PYTHON3:
-    intern = sys.intern
-    L65536 = 65536
-
-    def long(num):
-        return num
+L65536 = 65536
 
 
-else:
-    L65536 = long(65536)  # NOQA
+def long(num):
+    return num
 
 
-class Code(object):
+CONST_COLLECTIONS = ("CONST_LIST", "CONST_SET", "CONST_DICT", "CONST_MAP")
+
+
+class Code:
     """
     Class for representing code-objects.
 
@@ -90,34 +99,119 @@ class Code(object):
     the diassembled code is stored in the attribute '_tokens'.
     """
 
-    def __init__(self, co, scanner, classname=None):
+    def __init__(self, co, scanner, classname=None, show_asm=None):
+        # Full initialization is given below, but for linters
+        # well set up some initial values.
+        self.co_code = None  # Really either bytes for >= 3.0 and string in < 3.0
+
         for i in dir(co):
             if i.startswith("co_"):
                 setattr(self, i, getattr(co, i))
-        self._tokens, self._customize = scanner.ingest(co, classname)
+        self._tokens, self._customize = scanner.ingest(co, classname, show_asm=show_asm)
 
 
-class Scanner(object):
-    def __init__(self, version, show_asm=None, is_pypy=False):
+class Scanner(ABC):
+    def __init__(self, version: tuple, show_asm=None, is_pypy=False):
         self.version = version
         self.show_asm = show_asm
         self.is_pypy = is_pypy
 
-        if version in PYTHON_VERSIONS:
+        # Temporary initialization.
+        self.opc = ModuleType("uninitialized")
+
+        if version[:2] in PYTHON_VERSIONS:
+            v_str = f"""opcode_{version_tuple_to_str(version, start=0, end=2, delimiter="")}"""
+            module_name = f"xdis.opcodes.{v_str}"
             if is_pypy:
-                v_str = "opcode_%spypy" % (int(version * 10))
-            else:
-                v_str = "opcode_%s" % (int(version * 10))
-            exec("from xdis.opcodes import %s" % v_str)
-            exec("self.opc = %s" % v_str)
+                module_name += "pypy"
+            self.opc = importlib.import_module(module_name)
         else:
-            raise TypeError("%s is not a Python version I know about" % version)
+            raise TypeError(
+                "%s is not a Python version I know about"
+                % version_tuple_to_str(version)
+            )
 
         self.opname = self.opc.opname
 
         # FIXME: This weird Python2 behavior is not Python3
         self.resetTokenClass()
 
+    def bound_collection_from_tokens(self, tokens, t, i, collection_type):
+        count = t.attr
+        assert isinstance(count, int)
+
+        assert count <= i
+
+        if collection_type == "CONST_DICT":
+            # constant dictionaries work via BUILD_CONST_KEY_MAP and
+            # handle the values() like sets and lists.
+            # However, the keys() are an LOAD_CONST of the keys.
+            # adjust offset to account for this
+            count += 1
+
+        # For small lists don't bother
+        if count < 5:
+            return None
+
+        collection_start = i - count
+
+        for j in range(collection_start, i):
+            if tokens[j].kind not in (
+                "LOAD_CONST",
+                "LOAD_FAST",
+                "LOAD_GLOBAL",
+                "LOAD_NAME",
+            ):
+                return None
+
+        collection_enum = CONST_COLLECTIONS.index(collection_type)
+
+        # If we go there all instructions before tokens[i] are LOAD_CONST and we can replace
+        # add a boundary marker and change LOAD_CONST to something else
+        new_tokens = tokens[:-count]
+        start_offset = tokens[collection_start].offset
+        new_tokens.append(
+            Token(
+                opname="COLLECTION_START",
+                attr=collection_enum,
+                pattr=collection_type,
+                offset="%s_0" % start_offset,
+                has_arg=True,
+                opc=self.opc,
+                has_extended_arg=False,
+            )
+        )
+        for j in range(collection_start, i):
+            if tokens[j] == "LOAD_CONST":
+                opname = "ADD_VALUE"
+            else:
+                opname = "ADD_VALUE_VAR"
+            new_tokens.append(
+                Token(
+                    opname=opname,
+                    attr=tokens[j].attr,
+                    pattr=tokens[j].pattr,
+                    offset=tokens[j].offset,
+                    has_arg=True,
+                    linestart=tokens[j].linestart,
+                    opc=self.opc,
+                    has_extended_arg=False,
+                )
+            )
+        new_tokens.append(
+            Token(
+                opname="BUILD_%s" % collection_type,
+                attr=t.attr,
+                pattr=t.pattr,
+                offset=t.offset,
+                has_arg=t.has_arg,
+                linestart=t.linestart,
+                opc=t.opc,
+                has_extended_arg=False,
+            )
+        )
+        return new_tokens
+
     def build_instructions(self, co):
         """
         Create a list of instructions (a structured object rather than
@@ -144,11 +238,10 @@ class Scanner(object):
 
         # Offset: lineno pairs, only for offsets which start line.
         # Locally we use list for more convenient iteration using indices
-        if self.version > 1.4:
-            linestarts = list(self.opc.findlinestarts(code_obj))
-        else:
-            linestarts = [[0, 1]]
+        linestarts = list(self.opc.findlinestarts(code_obj))
         self.linestarts = dict(linestarts)
+        if not self.linestarts:
+            return []
 
         # 'List-map' which shows line number of current op and offset of
         # first op on following line, given offset of op as index
@@ -189,7 +282,7 @@ class Scanner(object):
             for _ in range(instruction_size(op, self.opc)):
                 self.prev_op.append(offset)
 
-    def is_jump_forward(self, offset):
+    def is_jump_forward(self, offset: int) -> bool:
         """
         Return True if the code at offset is some sort of jump forward.
         That is, it is ether "JUMP_FORWARD" or an absolute jump that
@@ -202,10 +295,16 @@ class Scanner(object):
             return False
         return offset < self.get_target(offset)
 
-    def prev_offset(self, offset):
+    def ingest(self, co, classname=None, code_objects={}, show_asm=None):
+        """
+        Code to tokenize disassembly. Subclasses must implement this.
+        """
+        raise NotImplementedError("This method should have been implemented")
+
+    def prev_offset(self, offset: int) -> int:
         return self.insts[self.offset2inst_index[offset] - 1].offset
 
-    def get_inst(self, offset):
+    def get_inst(self, offset: int):
         # Instructions can get moved as a result of EXTENDED_ARGS removal.
         # So if "offset" is not in self.offset2inst_index, then
         # we assume that it was an instruction moved back.
@@ -216,7 +315,7 @@ class Scanner(object):
             assert self.code[offset] == self.opc.EXTENDED_ARG
         return self.insts[self.offset2inst_index[offset]]
 
-    def get_target(self, offset, extended_arg=0):
+    def get_target(self, offset: int, extended_arg: int = 0) -> int:
         """
         Get next instruction offset for op located at given <offset>.
         NOTE: extended_arg is no longer used
@@ -229,23 +328,14 @@ class Scanner(object):
             target = next_offset(inst.opcode, self.opc, inst.offset)
         return target
 
-    def get_argument(self, pos):
+    def get_argument(self, pos: int):
         arg = self.code[pos + 1] + self.code[pos + 2] * 256
         return arg
 
-    def next_offset(self, op, offset):
+    def next_offset(self, op, offset: int) -> int:
         return xdis.next_offset(op, self.opc, offset)
 
-    def print_bytecode(self):
-        for i in self.op_range(0, len(self.code)):
-            op = self.code[i]
-            if op in self.JUMP_OPS:
-                dest = self.get_target(i, op)
-                print("%i\t%s\t%i" % (i, self.opname[op], dest))
-            else:
-                print("%i\t%s\t" % (i, self.opname[op]))
-
-    def first_instr(self, start, end, instr, target=None, exact=True):
+    def first_instr(self, start: int, end: int, instr, target=None, exact=True):
         """
         Find the first <instr> in the block from start to end.
         <instr> is any python bytecode instruction or a list of opcodes
@@ -279,7 +369,9 @@ class Scanner(object):
                         result_offset = offset
         return result_offset
 
-    def last_instr(self, start, end, instr, target=None, exact=True):
+    def last_instr(
+        self, start: int, end: int, instr, target=None, exact=True
+    ) -> Optional[int]:
         """
         Find the last <instr> in the block from start to end.
         <instr> is any python bytecode instruction or a list of opcodes
@@ -344,7 +436,7 @@ class Scanner(object):
         """
         try:
             None in instr
-        except:
+        except Exception:
             instr = [instr]
 
         first = self.offset2inst_index[start]
@@ -375,7 +467,9 @@ class Scanner(object):
     # FIXME: this is broken on 3.6+. Replace remaining (2.x-based) calls
     # with inst_matches
 
-    def all_instr(self, start, end, instr, target=None, include_beyond_target=False):
+    def all_instr(
+        self, start: int, end: int, instr, target=None, include_beyond_target=False
+    ):
         """
         Find all `instr` in the block from start to end.
         `instr` is any Python opcode or a list of opcodes
@@ -396,7 +490,6 @@ class Scanner(object):
         result = []
         extended_arg = 0
         for offset in self.op_range(start, end):
-
             op = code[offset]
 
             if op == self.opc.EXTENDED_ARG:
@@ -455,7 +548,6 @@ class Scanner(object):
                 offset = inst.offset
                 continue
             if last_was_extarg:
-
                 # j = self.stmts.index(inst.offset)
                 # self.lines[j] = offset
 
@@ -502,39 +594,36 @@ class Scanner(object):
     def resetTokenClass(self):
         return self.setTokenClass(Token)
 
-    def restrict_to_parent(self, target, parent):
+    def restrict_to_parent(self, target: int, parent) -> int:
         """Restrict target to parent structure boundaries."""
         if not (parent["start"] < target < parent["end"]):
             target = parent["end"]
         return target
 
-    def setTokenClass(self, tokenClass):
-        # assert isinstance(tokenClass, types.ClassType)
+    def setTokenClass(self, tokenClass: Token) -> Token:
         self.Token = tokenClass
         return self.Token
 
 
-def parse_fn_counts(argc):
-    return ((argc & 0xFF), (argc >> 8) & 0xFF, (argc >> 16) & 0x7FFF)
-
-
-def get_scanner(version, is_pypy=False, show_asm=None):
-
+def get_scanner(version: Union[str, tuple], is_pypy=False, show_asm=None) -> Scanner:
+    """
+    Import the right scanner module for ``version`` and return the Scanner class
+    in that module.
+    """
     # If version is a string, turn that into the corresponding float.
     if isinstance(version, str):
         if version not in canonic_python_version:
-            raise RuntimeError("Unknown Python version in xdis %s" % version)
+            raise RuntimeError(f"Unknown Python version in xdis {version}")
         canonic_version = canonic_python_version[version]
         if canonic_version not in CANONIC2VERSION:
             raise RuntimeError(
-                "Unsupported Python version %s (canonic %s)"
-                % (version, canonic_version)
+                f"Unsupported Python version {version} (canonic {canonic_version})"
             )
         version = CANONIC2VERSION[canonic_version]
 
     # Pick up appropriate scanner
-    if version in PYTHON_VERSIONS:
-        v_str = "%s" % (int(version * 10))
+    if version[:2] in PYTHON_VERSIONS:
+        v_str = version_tuple_to_str(version, start=0, end=2, delimiter="")
         try:
             import importlib
 
@@ -566,15 +655,19 @@ def get_scanner(version, is_pypy=False, show_asm=None):
                 "scan.Scanner%s(show_asm=show_asm)" % v_str, locals(), globals()
             )
     else:
-        raise RuntimeError("Unsupported Python version %s" % version)
+        raise RuntimeError(
+            f"Unsupported Python version, {version_tuple_to_str(version)}, for decompilation"
+        )
     return scanner
 
 
 if __name__ == "__main__":
-    import inspect, uncompyle6
+    import inspect
 
     co = inspect.currentframe().f_code
     # scanner = get_scanner('2.7.13', True)
     # scanner = get_scanner(sys.version[:5], False)
-    scanner = get_scanner(uncompyle6.PYTHON_VERSION, IS_PYPY, True)
+    from xdis.version_info import PYTHON_VERSION_TRIPLE
+
+    scanner = get_scanner(PYTHON_VERSION_TRIPLE, IS_PYPY, True)
     tokens, customize = scanner.ingest(co, {}, show_asm="after")
