@@ -17,23 +17,25 @@
 
 from uncompyle6.semantics.consts import TABLE_DIRECT
 
+
 #######################
 # Python 2.5+ Changes #
 #######################
 def customize_for_version25(self, version):
-
     ########################
     # Import style for 2.5+
     ########################
-    TABLE_DIRECT.update({
-        'importmultiple': ( '%|import %c%c\n', 2, 3 ),
-        'import_cont'   : ( ', %c', 2 ),
-        # With/as is allowed as "from future" thing in 2.5
-        # Note: It is safe to put the variables after "as" in parenthesis,
-        # and sometimes it is needed.
-        'with':     ( '%|with %c:\n%+%c%-', 0, 3),
-        'withasstmt':   ( '%|with %c as (%c):\n%+%c%-', 0, 2, 3),
-    })
+    self.TABLE_DIRECT.update(
+        {
+            "importmultiple": ("%|import %c%c\n", 2, 3),
+            "import_cont": (", %c", 2),
+            # With/as is allowed as "from future" thing in 2.5
+            # Note: It is safe to put the variables after "as" in parenthesis,
+            # and sometimes it is needed.
+            "with": ("%|with %c:\n%+%c%-", 0, 3),
+            "and_then": ("%c and %c", (0, "expr"), (4, "expr")),
+        }
+    )
 
     # In 2.5+ "except" handlers and the "finally" can appear in one
     # "try" statement. So the below has the effect of combining the
@@ -41,10 +43,18 @@ def customize_for_version25(self, version):
     # FIXME: something doesn't smell right, since the semantics
     # are different. See test_fileio.py for an example that shows this.
     def tryfinallystmt(node):
-        if len(node[1][0]) == 1 and node[1][0][0] == 'stmt':
-            if node[1][0][0][0] == 'try_except':
-                node[1][0][0][0].kind = 'tf_try_except'
-            if node[1][0][0][0] == 'tryelsestmt':
-                node[1][0][0][0].kind = 'tf_tryelsestmt'
+        if len(node[1][0]) == 1 and node[1][0][0] == "stmt":
+            if node[1][0][0][0] == "try_except":
+                node[1][0][0][0].kind = "tf_try_except"
+            if node[1][0][0][0] == "tryelsestmt":
+                node[1][0][0][0].kind = "tf_tryelsestmt"
         self.default(node)
+
     self.n_tryfinallystmt = tryfinallystmt
+
+    def n_import_from(node):
+        if node[0].pattr > 0:
+            node[2].pattr = ("." * node[0].pattr) + node[2].pattr
+        self.default(node)
+
+    self.n_import_from = n_import_from
