@@ -1,4 +1,4 @@
-#  Copyright (c) 2018-2020 by Rocky Bernstein
+#  Copyright (c) 2018-2021, 2023-2024 by Rocky Bernstein
 #
 #  This program is free software: you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -13,27 +13,24 @@
 #  You should have received a copy of the GNU General Public License
 #  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-"""Isolate Python 3 version-specific semantic actions here.
+"""
+Isolate Python 3 version-specific semantic actions here.
 """
 
+from xdis import iscode
+
 from uncompyle6.semantics.consts import TABLE_DIRECT
-
-from xdis import co_flags_is_async, iscode
-from uncompyle6.scanner import Code
-from uncompyle6.semantics.helper import (
-    find_code_node,
-    gen_function_parens_adjust,
-)
-
-from uncompyle6.semantics.make_function3 import make_function3_annotate
 from uncompyle6.semantics.customize35 import customize_for_version35
 from uncompyle6.semantics.customize36 import customize_for_version36
 from uncompyle6.semantics.customize37 import customize_for_version37
 from uncompyle6.semantics.customize38 import customize_for_version38
+from uncompyle6.semantics.helper import find_code_node, gen_function_parens_adjust
+from uncompyle6.semantics.make_function3 import make_function3_annotate
+from uncompyle6.util import get_code_name
 
 
-def customize_for_version3(self, version):
-    TABLE_DIRECT.update(
+def customize_for_version3(self, version: tuple):
+    self.TABLE_DIRECT.update(
         {
             "comp_for": (" for %c in %c", (2, "store"), (0, "expr")),
             "if_exp_not": (
@@ -51,14 +48,13 @@ def customize_for_version3(self, version):
             "import_cont": (", %c", 2),
             "kwarg": ("%[0]{attr}=%c", 1),
             "raise_stmt2": ("%|raise %c from %c\n", 0, 1),
-            "tf_tryelsestmtl3": ( '%c%-%c%|else:\n%+%c', 1, 3, 5 ),
+            "tf_tryelsestmtl3": ("%c%-%c%|else:\n%+%c", 1, 3, 5),
             "store_locals": ("%|# inspect.currentframe().f_locals = __locals__\n",),
             "with": ("%|with %c:\n%+%c%-", 0, 3),
-            "withasstmt": ("%|with %c as (%c):\n%+%c%-", 0, 2, 3),
         }
     )
 
-    assert version >= 3.0
+    assert version >= (3, 0)
 
     # In 2.5+ and 3.0+ "except" handlers and the "finally" can appear in one
     # "try" statement. So the below has the effect of combining the
@@ -67,147 +63,24 @@ def customize_for_version3(self, version):
     # are different. See test_fileio.py for an example that shows this.
     def tryfinallystmt(node):
         suite_stmts = node[1][0]
-        if len(suite_stmts) == 1 and suite_stmts[0] == 'stmt':
+        if len(suite_stmts) == 1 and suite_stmts[0] == "stmt":
             stmt = suite_stmts[0]
             try_something = stmt[0]
             if try_something == "try_except":
                 try_something.kind = "tf_try_except"
             if try_something.kind.startswith("tryelsestmt"):
                 if try_something == "tryelsestmtl3":
-                    try_something.kind = 'tf_tryelsestmtl3'
+                    try_something.kind = "tf_tryelsestmtl3"
                 else:
-                    try_something.kind = 'tf_tryelsestmt'
+                    try_something.kind = "tf_tryelsestmt"
         self.default(node)
+
     self.n_tryfinallystmt = tryfinallystmt
 
-    def listcomp_closure3(node):
-        """List comprehensions in Python 3 when handled as a closure.
-        See if we can combine code.
-        """
-        p = self.prec
-        self.prec = 27
-
-        code_obj = node[1].attr
-        assert iscode(code_obj)
-        code = Code(code_obj, self.scanner, self.currentclass)
-        ast = self.build_ast(code._tokens, code._customize, code)
-        self.customize(code._customize)
-
-        # skip over: sstmt, stmt, return, ret_expr
-        # and other singleton derivations
-        while len(ast) == 1 or (
-            ast in ("sstmt", "return") and ast[-1] in ("RETURN_LAST", "RETURN_VALUE")
-        ):
-            self.prec = 100
-            ast = ast[0]
-
-        n = ast[1]
-
-        # collections is the name of the expression(s) we are iterating over
-        collections = [node[-3]]
-        list_ifs = []
-
-        if self.version == 3.0 and n != "list_iter":
-            # FIXME 3.0 is a snowflake here. We need
-            # special code for this. Not sure if this is totally
-            # correct.
-            stores = [ast[3]]
-            assert ast[4] == "comp_iter"
-            n = ast[4]
-            # Find the list comprehension body. It is the inner-most
-            # node that is not comp_.. .
-            while n == "comp_iter":
-                if n[0] == "comp_for":
-                    n = n[0]
-                    stores.append(n[2])
-                    n = n[3]
-                elif n[0] in ("comp_if", "comp_if_not"):
-                    n = n[0]
-                    # FIXME: just a guess
-                    if n[0].kind == "expr":
-                        list_ifs.append(n)
-                    else:
-                        list_ifs.append([1])
-                    n = n[2]
-                    pass
-                else:
-                    break
-                pass
-
-            # Skip over n[0] which is something like: _[1]
-            self.preorder(n[1])
-
-        else:
-            assert n == "list_iter"
-            stores = []
-            # Find the list comprehension body. It is the inner-most
-            # node that is not list_.. .
-            while n == "list_iter":
-
-                # recurse one step
-                n = n[0]
-
-                if n == "list_for":
-                    stores.append(n[2])
-                    n = n[3]
-                    if n[0] == "list_for":
-                        # Dog-paddle down largely singleton reductions
-                        # to find the collection (expr)
-                        c = n[0][0]
-                        if c == "expr":
-                            c = c[0]
-                        # FIXME: grammar is wonky here? Is this really an attribute?
-                        if c == "attribute":
-                            c = c[0]
-                        collections.append(c)
-                        pass
-                elif n in ("list_if", "list_if_not"):
-                    # FIXME: just a guess
-                    if n[0].kind == "expr":
-                        list_ifs.append(n)
-                    else:
-                        list_ifs.append([1])
-                    n = n[2]
-                    pass
-                elif n == "list_if37":
-                    list_ifs.append(n)
-                    n = n[-1]
-                    pass
-                elif n == "list_afor":
-                    collections.append(n[0][0])
-                    n = n[1]
-                    stores.append(n[1][0])
-                    n = n[3]
-                pass
-
-            assert n == "lc_body", ast
-
-            self.preorder(n[0])
-
-        # FIXME: add indentation around "for"'s and "in"'s
-        n_colls = len(collections)
-        for i, store in enumerate(stores):
-            if i >= n_colls:
-                break
-            if collections[i] == "LOAD_DEREF"  and co_flags_is_async(code_obj.co_flags):
-                self.write(" async")
-                pass
-            self.write(" for ")
-            self.preorder(store)
-            self.write(" in ")
-            self.preorder(collections[i])
-            if i < len(list_ifs):
-                self.preorder(list_ifs[i])
-                pass
-            pass
-        self.prec = p
-    self.listcomp_closure3 = listcomp_closure3
-
     def n_classdef3(node):
-        """Handle "classdef" nonterminal for 3.0 >= version 3.0 <= 3.5
-        """
+        """Handle "classdef" nonterminal for 3.0 >= version 3.0 < 3.6"""
 
-        assert 3.0 <= self.version <= 3.5
+        assert (3, 0) <= self.version < (3, 6)
 
         # class definition ('class X(A,B,C):')
         cclass = self.currentclass
@@ -220,7 +93,7 @@ def customize_for_version3(self, version):
         # * subclass_code - the code for the subclass body
         subclass_info = None
         if node == "classdefdeco2":
-            if self.version <= 3.3:
+            if self.version < (3, 4):
                 class_name = node[2][0].attr
             else:
                 class_name = node[1][2].attr
@@ -233,7 +106,7 @@ def customize_for_version3(self, version):
         assert "mkfunc" == build_class[1]
         mkfunc = build_class[1]
         if mkfunc[0] in ("kwargs", "no_kwargs"):
-            if 3.0 <= self.version <= 3.2:
+            if (3, 0) <= self.version < (3, 3):
                 for n in mkfunc:
                     if hasattr(n, "attr") and iscode(n.attr):
                         subclass_code = n.attr
@@ -265,14 +138,10 @@ def customize_for_version3(self, version):
                 # Python 3.2 works like this
                 subclass_code = find_code_node(load_closure, -2).attr
             else:
-                raise "Internal Error n_classdef: cannot find class body"
-            if hasattr(build_class[3], "__len__"):
-                if not subclass_info:
-                    subclass_info = build_class[3]
-            elif hasattr(build_class[2], "__len__"):
-                subclass_info = build_class[2]
-            else:
-                raise "Internal Error n_classdef: cannot superclass name"
+                raise RuntimeError("Internal Error n_classdef: cannot find class body")
+
+            subclass_info = build_class
+
         elif not subclass_info:
             if mkfunc[0] in ("no_kwargs", "kwargs"):
                 subclass_code = mkfunc[1].attr
@@ -314,18 +183,25 @@ def customize_for_version3(self, version):
         # the iteration variable. These rules we can ignore
         # since we pick up the iteration variable some other way and
         # we definitely don't include in the source  _[dd].
-        TABLE_DIRECT.update({
-            "ifstmt30":	( "%|if %c:\n%+%c%-",
-                          (0, "testfalse_then"),
-                          (1, "_ifstmts_jump30") ),
-            "ifnotstmt30": ( "%|if not %c:\n%+%c%-",
-                             (0, "testtrue_then"),
-                             (1, "_ifstmts_jump30") ),
-            "try_except30": ( "%|try:\n%+%c%-%c\n\n",
-                              (1, "suite_stmts_opt"),
-                              (4, "except_handler") ),
-
-            })
+        self.TABLE_DIRECT.update(
+            {
+                "ifstmt30": (
+                    "%|if %c:\n%+%c%-",
+                    (0, "testfalse_then"),
+                    (1, "_ifstmts_jump30"),
+                ),
+                "ifnotstmt30": (
+                    "%|if not %c:\n%+%c%-",
+                    (0, "testtrue_then"),
+                    (1, "_ifstmts_jump30"),
+                ),
+                "try_except30": (
+                    "%|try:\n%+%c%-%c\n\n",
+                    (1, "suite_stmts_opt"),
+                    (4, "except_handler"),
+                ),
+            }
+        )
 
         def n_comp_iter(node):
             if node[0] == "expr":
@@ -345,7 +221,7 @@ def customize_for_version3(self, version):
             assert node[0] == "expr"
             if node[0][0] == "get_iter":
                 # Skip over yield_from.expr.get_iter which adds an
-                # extra iter(). Maybe we can do in tranformation phase instead?
+                # extra iter(). Maybe we can do in transformation phase instead?
                 template = ("yield from %c", (0, "expr"))
                 self.template_engine(template, node[0][0])
             else:
@@ -355,10 +231,9 @@ def customize_for_version3(self, version):
 
         self.n_yield_from = n_yield_from
 
-    if 3.2 <= version <= 3.4:
+    if (3, 2) <= version <= (3, 4):
 
         def n_call(node):
-
             mapping = self._get_mapping(node)
             key = node
             for i in mapping[1:]:
@@ -398,7 +273,7 @@ def customize_for_version3(self, version):
             self.default(node)
 
         self.n_call = n_call
-    elif version < 3.2:
+    elif version < (3, 2):
 
         def n_call(node):
             mapping = self._get_mapping(node)
@@ -412,24 +287,23 @@ def customize_for_version3(self, version):
         self.n_call = n_call
 
     def n_mkfunc_annotate(node):
-
         # Handling EXTENDED_ARG before MAKE_FUNCTION ...
         i = -1 if node[-2] == "EXTENDED_ARG" else 0
 
-        if self.version <= 3.2:
-            code = node[-2 + i]
-        elif self.version >= 3.3 or node[-2] == "kwargs":
+        if self.version < (3, 3):
+            code_node = node[-2 + i]
+        elif self.version >= (3, 3) or node[-2] == "kwargs":
             # LOAD_CONST code object ..
             # LOAD_CONST        'x0'  if >= 3.3
             # EXTENDED_ARG
             # MAKE_FUNCTION ..
-            code = node[-3 + i]
+            code_node = node[-3 + i]
         elif node[-3] == "expr":
-            code = node[-3][0]
+            code_node = node[-3][0]
         else:
             # LOAD_CONST code object ..
             # MAKE_FUNCTION ..
-            code = node[-3]
+            code_node = node[-3]
 
         self.indent_more()
         for annotate_last in range(len(node) - 1, -1, -1):
@@ -439,13 +313,17 @@ def customize_for_version3(self, version):
         # FIXME: the real situation is that when derived from
         # function_def_annotate we the name has been filled in.
         # But when derived from funcdefdeco it hasn't Would like a better
-        # way to distinquish.
+        # way to distinguish.
         if self.f.getvalue()[-4:] == "def ":
-            self.write(code.attr.co_name)
+            self.write(get_code_name(code_node.attr))
 
         # FIXME: handle and pass full annotate args
         make_function3_annotate(
-            self, node, is_lambda=False, code_node=code, annotate_last=annotate_last
+            self,
+            node,
+            is_lambda=False,
+            code_node=code_node,
+            annotate_last=annotate_last,
         )
 
         if len(self.param_stack) > 1:
@@ -457,34 +335,34 @@ def customize_for_version3(self, version):
 
     self.n_mkfunc_annotate = n_mkfunc_annotate
 
-    TABLE_DIRECT.update(
+    self.TABLE_DIRECT.update(
         {
             "tryelsestmtl3": (
                 "%|try:\n%+%c%-%c%|else:\n%+%c%-",
                 (1, "suite_stmts_opt"),
-                3, # "except_handler_else" or "except_handler"
+                3,  # "except_handler_else" or "except_handler"
                 (5, "else_suitel"),
             ),
             "LOAD_CLASSDEREF": ("%{pattr}",),
         }
     )
-    if version >= 3.4:
+    if version >= (3, 4):
         #######################
         # Python 3.4+ Changes #
         #######################
-        TABLE_DIRECT.update(
+        self.TABLE_DIRECT.update(
             {
                 "LOAD_CLASSDEREF": ("%{pattr}",),
                 "yield_from": ("yield from %c", (0, "expr")),
             }
         )
-        if version >= 3.5:
+        if version >= (3, 5):
             customize_for_version35(self, version)
-            if version >= 3.6:
+            if version >= (3, 6):
                 customize_for_version36(self, version)
-                if version >= 3.7:
+                if version >= (3, 7):
                     customize_for_version37(self, version)
-                    if version >= 3.8:
+                    if version >= (3, 8):
                         customize_for_version38(self, version)
                         pass  # version >= 3.8
                     pass  # 3.7
