@@ -1,4 +1,4 @@
-#  Copyright (c) 2019-2020 by Rocky Bernstein
+#  Copyright (c) 2019-2024 by Rocky Bernstein
 #
 #  This program is free software: you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -15,16 +15,18 @@
 """Isolate Python 3.6 version-specific semantic actions here.
 """
 
-from xdis.code import iscode
 from spark_parser.ast import GenericASTTraversalPruningException
+from xdis import iscode
+
 from uncompyle6.scanners.tok import Token
-from uncompyle6.semantics.helper import flatten_list, escape_string, strip_quotes
 from uncompyle6.semantics.consts import (
     INDENT_PER_LEVEL,
     PRECEDENCE,
     TABLE_DIRECT,
     TABLE_R,
 )
+from uncompyle6.semantics.helper import escape_string, flatten_list, strip_quotes
+from uncompyle6.util import get_code_name
 
 
 def escape_format(s):
@@ -36,53 +38,72 @@ def escape_format(s):
 #######################
 
 
-def customize_for_version36(self, version):
-    PRECEDENCE["call_kw"] = 0
-    PRECEDENCE["call_kw36"] = 1
-    PRECEDENCE["call_ex"] = 1
-    PRECEDENCE["call_ex_kw"] = 1
-    PRECEDENCE["call_ex_kw2"] = 1
-    PRECEDENCE["call_ex_kw3"] = 1
-    PRECEDENCE["call_ex_kw4"] = 1
-    PRECEDENCE["unmap_dict"] = 0
+def customize_for_version36(self, version: tuple):
+    # fmt: off
+    PRECEDENCE["call_kw"]          =   0
+    PRECEDENCE["call_kw36"]        =   1
+    PRECEDENCE["call_ex"]          =   1
+    PRECEDENCE["call_ex_kw"]       =   1
+    PRECEDENCE["call_ex_kw2"]      =   1
+    PRECEDENCE["call_ex_kw3"]      =   1
+    PRECEDENCE["call_ex_kw4"]      =   1
+    PRECEDENCE["dict_pack"]        =   0  # **{ ... }
     PRECEDENCE["formatted_value1"] = 100
 
-    TABLE_DIRECT.update(
+    self.TABLE_DIRECT.update(
         {
             "ann_assign_init_value": (
                 "%|%c = %p\n",
                 (-1, "store_annotation"),
-                (0, "expr", 200),
+                (0,  "expr", 200),
             ),
             "ann_assign_no_init": ("%|%c\n", (0, "store_annotation")),
             "async_for_stmt36": (
                 "%|async for %c in %c:\n%+%c%-\n\n",
                 (9, "store"),
                 (1, "expr"),
-                (18, "for_block"),
+                # Count from end, since COME_FROM shifts things in the forward direction
+                (-9, ("for_block", "pass")),
             ),
-            "call_ex": ("%c(%p)", (0, "expr"), (1, 100)),
-            "except_return": ("%|except:\n%+%c%-", 3),
-            "func_args36": ("%c(**", 0),
+            "async_forelse_stmt36": (
+                "%|async for %c in %c:\n%+%c%-%|else:\n%+%c%-\n\n",
+                (9, "store"),
+                (1, "expr"),
+                (-10, "for_block"),
+                (-2, "else_suite"),
+            ),
+            "call_ex":            ("%c(%p)", (0, "expr"), (1, 100)),
+            "except_return":      ("%|except:\n%+%c%-", 3),
+            "func_args36":        ("%c(**", 0),
+
             # This comes from 3.7. Eventually we will rebase from 3.7
             # and then this can go away
-            "if_exp37": ("%p if %c else %c", (1, "expr", 27), 0, 3),
-            "ifstmtl": ("%|if %c:\n%+%c%-", (0, "testexpr"), (1, "_ifstmts_jumpl")),
-            "try_except36": ("%|try:\n%+%c%-%c\n\n", 1, -2),
-            "tryfinally36": ("%|try:\n%+%c%-%|finally:\n%+%c%-\n\n", (1, "returns"), 3),
+            "if_exp37":           ("%p if %c else %c", (1, "expr", 27), 0, 3),
+
+            "ifstmtl":            ("%|if %c:\n%+%c%-",
+                                   (0, "testexpr"), (1, "_ifstmts_jumpl")),
+
+            "list_afor": (
+                " async for %[1]{%c} in %c%[1]{%c}",
+                (1, "store"), (0, "get_aiter"), (3, "list_iter"),
+            ),
+
+            "try_except36":       ("%|try:\n%+%c%-%c\n\n", 1, -2),
+            "tryfinally36":       ("%|try:\n%+%c%-%|finally:\n%+%c%-\n\n", (1, "returns"), 3),
             "tryfinally_return_stmt": ("%|try:\n%+%c%-%|finally:\n%+%|return%-\n\n", 1),
-            "unpack_list": ("*%c", (0, "list")),
-            "store_annotation": ("%[1]{pattr}: %c", 0),
+            "unpack_list":        ("*%c", (0, "list")),
+            "store_annotation":   ("%[1]{pattr}: %c", 0),
         }
     )
 
-    TABLE_R.update(
+    self.TABLE_R.update(
         {
-            "CALL_FUNCTION_EX": ("%c(*%P)", 0, (1, 2, ", ", 100)),
+            "CALL_FUNCTION_EX":    ("%c(*%P)", 0, (1, 2, ", ", 100)),
             # Not quite right
             "CALL_FUNCTION_EX_KW": ("%c(**%C)", 0, (2, 3, ",")),
         }
     )
+    # fmt: on
 
     def build_unpack_tuple_with_call(node):
         n = node[0]
@@ -148,6 +169,9 @@ def customize_for_version36(self, version):
         if node == "classdefdeco2":
             if isinstance(node[1][1].attr, str):
                 class_name = node[1][1].attr
+                if self.is_pypy and class_name.find("<locals>") > 0:
+                    class_name = class_name.split(".")[-1]
+
             else:
                 class_name = node[1][2].attr
             build_class = node
@@ -167,7 +191,7 @@ def customize_for_version36(self, version):
                 code_node = build_class[1][1]
             else:
                 code_node = build_class[1][0]
-            class_name = code_node.attr.co_name
+            class_name = get_code_name(code_node.attr)
 
         assert "mkfunc" == build_class[1]
         mkfunc = build_class[1]
@@ -184,23 +208,24 @@ def customize_for_version36(self, version):
         elif build_class[1][0] == "load_closure":
             # Python 3 with closures not functions
             load_closure = build_class[1]
-            if hasattr(load_closure[-3], "attr"):
-                # Python 3.3 classes with closures work like this.
-                # Note have to test before 3.2 case because
-                # index -2 also has an attr.
-                subclass_code = load_closure[-3].attr
-            elif hasattr(load_closure[-2], "attr"):
-                # Python 3.2 works like this
-                subclass_code = load_closure[-2].attr
-            else:
-                raise "Internal Error n_classdef: cannot find class body"
+            subclass_code = None
+            for i in range(-4, -1):
+                if load_closure[i] == "LOAD_CODE":
+                    subclass_code = load_closure[i].attr
+                    break
+            if subclass_code is None:
+                raise RuntimeError(
+                    "Internal Error n_classdef: cannot find " "class body"
+                )
             if hasattr(build_class[3], "__len__"):
                 if not subclass_info:
                     subclass_info = build_class[3]
             elif hasattr(build_class[2], "__len__"):
                 subclass_info = build_class[2]
             else:
-                raise "Internal Error n_classdef: cannot superclass name"
+                raise RuntimeError(
+                    "Internal Error n_classdef: cannot " "superclass name"
+                )
         elif node == "classdefdeco2":
             subclass_info = node
             subclass_code = build_class[1][0].attr
@@ -251,7 +276,7 @@ def customize_for_version36(self, version):
         if value == "":
             fmt = "%c(%p)"
         else:
-            fmt = "%%c(%s, %%p)" % value
+            fmt = "%c" + ("(%s, " % value).replace("%", "%%") + "%p)"
 
         self.template_engine(
             (fmt, (0, "expr"), (2, "build_map_unpack_with_call", 100)), node
@@ -270,7 +295,7 @@ def customize_for_version36(self, version):
         if value == "":
             fmt = "%c(%p)"
         else:
-            fmt = "%%c(%s, %%p)" % value
+            fmt = "%c" + ("(%s, " % value).replace("%", "%%") + "%p)"
 
         self.template_engine(
             (fmt, (0, "expr"), (2, "build_map_unpack_with_call", 100)), node
@@ -332,7 +357,7 @@ def customize_for_version36(self, version):
             kwargs = kwargs[0]
         call_function_ex = node[-1]
         assert call_function_ex == "CALL_FUNCTION_EX_KW" or (
-            self.version >= 3.6 and call_function_ex == "CALL_FUNCTION_EX"
+            self.version >= (3, 6) and call_function_ex == "CALL_FUNCTION_EX"
         )
         # FIXME: decide if the below test be on kwargs == 'dict'
         if (
@@ -374,7 +399,7 @@ def customize_for_version36(self, version):
 
     def call36_tuple(node):
         """
-        A tuple used in a call, these are like normal tuples but they
+        A tuple used in a call; these are like normal tuples, but they
         don't have the enclosing parenthesis.
         """
         assert node == "tuple"
@@ -649,7 +674,7 @@ def customize_for_version36(self, version):
             else:
                 # {{ and }} in Python source-code format strings mean
                 # { and } respectively. But only when *not* part of a
-                # formatted value. However in the LOAD_STR
+                # formatted value. However, in the LOAD_STR
                 # bytecode, the escaping of the braces has been
                 # removed. So we need to put back the braces escaping in
                 # reconstructing the source.
@@ -674,6 +699,17 @@ def customize_for_version36(self, version):
 
     self.n_joined_str = n_joined_str
 
+    def n_list_comp_async(node):
+        self.write("[")
+        if node[0].kind == "load_closure":
+            self.listcomp_closure3(node)
+        else:
+            self.comprehension_walk_newer(node, iter_index=3, code_index=0)
+        self.write("]")
+        self.prune()
+
+    self.n_list_comp_async = n_list_comp_async
+
     # def kwargs_only_36(node):
     #     keys = node[-1].attr
     #     num_kwargs = len(keys)
